/*
Copyright (C) 2014 Graham Jones, www.indriid.com
*/
package stacey;

import beast.core.Description;
import beast.core.Input;
import beast.core.Operator;
import beast.evolution.tree.Tree;
import stacey.debugtune.Checks;
import stacey.util.*;

import java.io.PrintStream;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.List;

/**
 * Created by Graham Jones on 27/10/2014.
 */

@Description("A move which does a subtree-prune-regraft move on a smcTree node," +
        "then carries a coordinated set of similar moves on the gene trees to maintain compatibility. <br/>")

// This is a multiple-trees operator. Is it a TreeOperator?
public class CoordinatedPruneRegraft extends Operator {

    @SuppressWarnings({"CanBeFinal", "WeakerAccess"})
    public Input<Tree> smcTreeInput =
            new Input<Tree>("smcTree",
                    "The species tree or minimal clusters tree", Input.Validate.REQUIRED);

    @SuppressWarnings({"CanBeFinal", "WeakerAccess"})
    public Input<List<Tree>> geneTreesInput =
            new Input<List<Tree>>("geneTree",
                    "All gene trees",
                    new ArrayList<Tree>());


    private OperatorsSMCTree smcTree;
    private OperatorsGeneTree[] geneTrees;
    private Bindings bindings;


    private final boolean debugFlag = Boolean.valueOf(System.getProperty("stacey.debug"));
    private final int debugMaxBranches = 100;
    private int [] debugRejectCounts;
    private int [] debugAcceptCounts;
    private int debugLastNofBranches;






    @Override
    public void initAndValidate() throws Exception {
        //super.initAndValidate(); // ??
        /*if (smcTreeInput.get().getLeafNodeCount() < 3) {
            throw new Exception("CoordinatedPruneRegraft cannot be used if there are less than 3 minimal clusters.");
        }
        TODO: this upsets Beauti */
        bindings = new Bindings(smcTreeInput.get().getTaxonset());
        smcTree = new OperatorsSMCTree(smcTreeInput.get(), bindings);
        geneTrees = new OperatorsGeneTree[geneTreesInput.get().size()];
        for (int i = 0; i < geneTreesInput.get().size(); i++) {
            geneTrees[i] = new OperatorsGeneTree(geneTreesInput.get().get(i), bindings);
        }

        if (debugFlag) {
            debugRejectCounts = new int[debugMaxBranches +1];
            debugAcceptCounts = new int[debugMaxBranches +1];
        }

    }



    @Override
    public double proposal() {

        if (smcTreeInput.get().getLeafNodeCount() < 3) {
            return Double.NEGATIVE_INFINITY;
        }

        update();
        if (debugFlag) {
            assert Checks.debugAllCompatible(smcTree, geneTrees);
        }

        double logHR = doCoordinatedPruneRegraft(false);
        // TODO: choose NNI and not-NNI at random(?)
        reset();

        if (debugFlag) {
            update();
            assert Checks.debugAllCompatible(smcTree, geneTrees);
            reset();
        }

        return logHR;
    }



    private void update() {
        smcTree.update();
        for (int j = 0; j < geneTrees.length; j++) {
            geneTrees[j].update();
        }
    }


    private void reset() {
        smcTree.reset();
        for (int j = 0; j < geneTrees.length; j++) {
            geneTrees[j].reset();
        }
    }



    @Override
    public void accept() {
        super.accept();
        if (debugFlag) {
            if (debugLastNofBranches < debugMaxBranches) {
                debugAcceptCounts[debugLastNofBranches]++;
            }
        }
    }


    @Override
    public void reject(final int reason) {
        super.reject(reason);
        if (debugFlag) {
            if (debugLastNofBranches < debugMaxBranches) {
                debugRejectCounts[debugLastNofBranches]++;
            }
        }
    }


    @Override
    public void storeToFile(final PrintWriter out) {
        super.storeToFile(out);
        if (debugFlag) {
            out.print("{id:\"" + getID() + "\"");
            for (int nb = 0; nb < debugMaxBranches; nb++) {
                out.print(" " + nb +
                                ", accept:" + debugAcceptCounts[nb] +
                                ", reject:" + debugRejectCounts[nb]
                );
            }
            out.print("}");
        }
    }



    // choose smc tree move s->d, remember s,x for reverse move, get information about
    // the move (smi) to pass to gene trees, do the smc tree move, do the gene tree
    // moves, get information about the reverse move (revsmi), calculate the
    // HR for reverse move.
    private double doCoordinatedPruneRegraft(boolean NNI) {

        // TODO This is a fix for beast.core.State$Trie memory
        for (int j = 0; j < geneTrees.length; j++) {
            geneTreesInput.get().get(j).startEditing(null);
        }

        int d = -1;
        int s = -1;
        s = smcTree.chooseSubtreeForSPR(NNI);
        Integer [] debugnb = new Integer[1];
        d = smcTree.chooseDestinationForSPR(s, NNI, debugnb);
        debugLastNofBranches = debugnb[0];
        int x = smcTree.siblingOf(s);
        // Track where s and x end up by storing their unions (values not references!)
        BitUnion sUnion = new BitUnion(bindings.smcTreeTipCount());
        BitUnion xUnion = new BitUnion(bindings.smcTreeTipCount());
        sUnion.replaceWith(smcTree.getUnion(s));
        xUnion.replaceWith(smcTree.getUnion(x));
        // the move
        double logHR = 0.0;
        OperatorsSMCTree.SPRmoveInfo smi = smcTree.makeSPRmoveInfo(s, d);
        smcTree.doSPRmove(s, d);
        logHR += doCoordinatedSPRMoves(smi);
        // update the smcTree and gene trees, find s and x in new smcTree
        reset();
        update();
        int sNew = smcTree.indexFromUnion(sUnion);
        int xNew = smcTree.indexFromUnion(xUnion);
        // calculate the reverse HR
        OperatorsSMCTree.SPRmoveInfo revsmi = smcTree.makeSPRmoveInfo(sNew, xNew);
        logHR -= HRForCoordinatedSPRMoves(revsmi);

        return logHR;
    }



    // Carries out a set of SPR moves on each gene tree based on the species tree SPR move
    // which has just been done. Also finds the Hastings ratio for the move.
    // This is log(number of choices) for the move.
    double doCoordinatedSPRMoves(OperatorsSMCTree.SPRmoveInfo sppSPRInfo) {
        double logHR = 0;
        for (int j = 0; j < geneTrees.length; j++) {
            logHR += geneTrees[j].doCoordinatedSPRMoves(sppSPRInfo);
        }
        return  logHR;
    }


    // For calculation of Hastings ratio for SPR move. This finds log(number of choices)
    // for the reverse move.
    double HRForCoordinatedSPRMoves(OperatorsSMCTree.SPRmoveInfo sppSPRInfo) {
        double logHR = 0;
        for (int j = 0; j < geneTrees.length; j++) {
            logHR += geneTrees[j].HRForCoordinatedSPRMoves(sppSPRInfo);
        }
        return  logHR;
    }





}
